"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DocumentContextManager = void 0;
const assert_1 = __importDefault(require("assert"));
const events_1 = require("events");
const documentContext_1 = require("./documentContext");
const LastCheckpointedOffset = {
    offset: -1,
    partition: -1,
    topic: "",
    value: undefined,
};
/**
 * The DocumentContextManager manages a set of created DocumentContexts and computes an aggregate checkpoint offset
 * from them.
 */
class DocumentContextManager extends events_1.EventEmitter {
    constructor(partitionContext) {
        super();
        this.partitionContext = partitionContext;
        this.contexts = new Set();
        // Head and tail represent our processing position of the queue. Head is the latest message seen and
        // tail is the last message processed
        this.head = LastCheckpointedOffset;
        this.tail = LastCheckpointedOffset;
        // Offset represents the last offset checkpointed
        this.lastCheckpoint = LastCheckpointedOffset;
        this.closed = false;
    }
    /**
     * Creates a context that should be used for a single document partition
     * This class is responsible for the lifetime of the context
     */
    createContext(routingKey, head) {
        // Contexts should only be created within the processing range of the manager
        (0, assert_1.default)(head.offset > this.tail.offset && head.offset <= this.head.offset);
        // Create the new context and register for listeners on it
        const context = new documentContext_1.DocumentContext(routingKey, head, this.partitionContext.log, () => this.tail);
        this.contexts.add(context);
        context.addListener("checkpoint", () => this.updateCheckpoint());
        context.addListener("error", (error, errorData) => this.emit("error", error, errorData));
        return context;
    }
    removeContext(context) {
        context.close();
        this.contexts.delete(context);
    }
    getHeadOffset() {
        return this.head.offset;
    }
    /**
     * Updates the head to the new offset. The head offset will not be updated if it stays the same or moves backwards.
     * @returns True if the head was updated, false if it was not.
     */
    setHead(head) {
        if (head.offset > this.head.offset) {
            this.head = head;
            return true;
        }
        return false;
    }
    setTail(tail) {
        (0, assert_1.default)(tail.offset > this.tail.offset && tail.offset <= this.head.offset, `${tail.offset} > ${this.tail.offset} && ${tail.offset} <= ${this.head.offset}`);
        this.tail = tail;
        this.updateCheckpoint();
    }
    close() {
        this.closed = true;
        for (const context of this.contexts) {
            context.close();
        }
        this.contexts.clear();
        this.removeAllListeners();
    }
    updateCheckpoint() {
        if (this.closed) {
            return;
        }
        // Set the starting offset at the tail. Contexts can then lower that offset based on their positions.
        let queuedMessage = this.tail;
        for (const context of this.contexts) {
            // Utilize the tail of the context if there is still pending work.
            // If there isn't pending work then we are fully caught up.
            if (context.hasPendingWork()) {
                // Lower the offset when possible
                queuedMessage = queuedMessage.offset > context.tail.offset ? context.tail : queuedMessage;
            }
        }
        // Checkpoint once the offset has changed
        if (queuedMessage.offset !== this.lastCheckpoint.offset) {
            this.partitionContext.checkpoint(queuedMessage);
            this.lastCheckpoint = queuedMessage;
        }
    }
}
exports.DocumentContextManager = DocumentContextManager;
//# sourceMappingURL=contextManager.js.map