"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DocumentContext = void 0;
const assert_1 = __importDefault(require("assert"));
const events_1 = require("events");
class DocumentContext extends events_1.EventEmitter {
    constructor(routingKey, head, log, getLatestTail) {
        super();
        this.routingKey = routingKey;
        this.log = log;
        this.getLatestTail = getLatestTail;
        this.closed = false;
        this.contextError = undefined;
        // Head represents the largest offset related to the document that is not checkpointed.
        // Tail will be set to the checkpoint offset of the previous head
        this.headInternal = head;
        this.tailInternal = this.getLatestTail();
    }
    get head() {
        return this.headInternal;
    }
    get tail() {
        return this.tailInternal;
    }
    /**
     * Returns whether or not there is pending work in flight - i.e. the head and tail are not equal
     */
    hasPendingWork() {
        return this.headInternal !== this.tailInternal;
    }
    /**
     * Updates the head offset for the context.
     */
    setHead(head) {
        (0, assert_1.default)(head.offset > this.head.offset, `${head.offset} > ${this.head.offset} ` +
            `(${head.topic}, ${head.partition}, ${this.routingKey.tenantId}/${this.routingKey.documentId})`);
        // When moving back to a state where head and tail differ we set the tail to be the old head, as in the
        // constructor, to make tail represent the inclusive top end of the checkpoint range.
        if (!this.hasPendingWork()) {
            this.tailInternal = this.getLatestTail();
        }
        this.headInternal = head;
    }
    checkpoint(message) {
        if (this.closed) {
            return;
        }
        // Assert offset is between the current tail and head
        const offset = message.offset;
        (0, assert_1.default)(offset > this.tail.offset && offset <= this.head.offset, `${offset} > ${this.tail.offset} && ${offset} <= ${this.head.offset} ` +
            `(${message.topic}, ${message.partition}, ${this.routingKey.tenantId}/${this.routingKey.documentId})`);
        // Update the tail and broadcast the checkpoint
        this.tailInternal = message;
        this.emit("checkpoint", this);
    }
    error(error, errorData) {
        this.contextError = error;
        this.emit("error", error, errorData);
    }
    close() {
        this.closed = true;
        this.removeAllListeners();
    }
    getContextError() {
        return this.contextError;
    }
}
exports.DocumentContext = DocumentContext;
//# sourceMappingURL=documentContext.js.map