"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.DocumentLambda = void 0;
const server_services_core_1 = require("@fluidframework/server-services-core");
const contextManager_1 = require("./contextManager");
const documentPartition_1 = require("./documentPartition");
class DocumentLambda {
    constructor(factory, config, context, documentLambdaServerConfiguration) {
        this.factory = factory;
        this.config = config;
        this.context = context;
        this.documentLambdaServerConfiguration = documentLambdaServerConfiguration;
        this.documents = new Map();
        this.contextManager = new contextManager_1.DocumentContextManager(context);
        this.contextManager.on("error", (error, errorData) => {
            context.error(error, errorData);
        });
        this.activityCheckTimer = setInterval(this.inactivityCheck.bind(this), documentLambdaServerConfiguration.partitionActivityCheckInterval);
    }
    handler(message) {
        var _a;
        if (!this.contextManager.setHead(message)) {
            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.warn("Unexpected head offset. " +
                `head offset: ${this.contextManager.getHeadOffset()}, message offset: ${message.offset}`);
            return undefined;
        }
        this.handlerCore(message);
        this.contextManager.setTail(message);
        return undefined;
    }
    close(closeType) {
        if (this.activityCheckTimer !== undefined) {
            clearInterval(this.activityCheckTimer);
            this.activityCheckTimer = undefined;
        }
        this.contextManager.close();
        for (const [, partition] of this.documents) {
            partition.close(closeType);
        }
        this.documents.clear();
    }
    handlerCore(message) {
        const boxcar = (0, server_services_core_1.extractBoxcar)(message);
        if (!boxcar.documentId || !boxcar.tenantId) {
            return;
        }
        // Stash the parsed value for down stream lambdas
        message.value = boxcar;
        // Create the routing key from tenantId + documentId
        const routingKey = `${boxcar.tenantId}/${boxcar.documentId}`;
        // Create or update the DocumentPartition
        let document = this.documents.get(routingKey);
        if (!document) {
            // Create a new context and begin tracking it
            const documentContext = this.contextManager.createContext(boxcar, message);
            document = new documentPartition_1.DocumentPartition(this.factory, this.config, boxcar.tenantId, boxcar.documentId, documentContext, this.documentLambdaServerConfiguration.partitionActivityTimeout);
            this.documents.set(routingKey, document);
        }
        else {
            // SetHead assumes it will always receive increasing offsets. So we need to split the creation case
            // from the update case.
            document.context.setHead(message);
        }
        // Forward the message to the document queue and then resolve the promise to begin processing more messages
        document.process(message);
    }
    /**
     * Closes inactive documents
     */
    inactivityCheck() {
        const now = Date.now();
        const documentPartitions = Array.from(this.documents);
        for (const [routingKey, documentPartition] of documentPartitions) {
            if (documentPartition.isInactive(now)) {
                // Close and remove the inactive document
                this.contextManager.removeContext(documentPartition.context);
                documentPartition.close(server_services_core_1.LambdaCloseType.ActivityTimeout);
                this.documents.delete(routingKey);
            }
        }
    }
}
exports.DocumentLambda = DocumentLambda;
//# sourceMappingURL=documentLambda.js.map