"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.DocumentPartition = void 0;
const util_1 = require("util");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const async_1 = require("async");
class DocumentPartition {
    constructor(factory, config, tenantId, documentId, context, activityTimeout) {
        this.tenantId = tenantId;
        this.documentId = documentId;
        this.context = context;
        this.activityTimeout = activityTimeout;
        this.corrupt = false;
        this.closed = false;
        this.updateActivityTime();
        const documentConfig = {
            leaderEpoch: config.leaderEpoch,
            tenantId,
            documentId,
        };
        this.q = (0, async_1.queue)((message, callback) => {
            // Winston.verbose(`${message.topic}:${message.partition}@${message.offset}`);
            try {
                if (!this.corrupt) {
                    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                    const optionalPromise = this.lambda.handler(message);
                    if (optionalPromise) {
                        optionalPromise
                            .then(callback)
                            .catch((error) => {
                            this.markAsCorrupt(error, message);
                            callback();
                        });
                        return;
                    }
                }
                else {
                    // Until we can dead letter - simply checkpoint as handled
                    this.context.checkpoint(message);
                }
            }
            catch (error) {
                // TODO dead letter queue for bad messages, etc... when the lambda is throwing an exception
                // for now we will simply continue on to keep the queue flowing
                this.markAsCorrupt(error, message);
            }
            // Handle the next message
            callback();
        }, 1);
        this.q.pause();
        this.context.on("error", (error, errorData) => {
            if (errorData.restart) {
                // ensure no more messages are processed by this partition
                // while the process is restarting / closing
                this.close(server_services_core_1.LambdaCloseType.Error);
            }
        });
        // Create the lambda to handle the document messages
        this.lambdaP = factory.create(documentConfig, context, this.updateActivityTime.bind(this));
        this.lambdaP.then((lambda) => {
            this.lambda = lambda;
            this.q.resume();
        }, (error) => {
            // There is no need to pass the message to be checkpointed to markAsCorrupt().
            // The message, in this case, would be the head in the DocumentContext. But the DocumentLambda
            // that creates this DocumentPartition will also put the same message in the queue.
            // So the DocumentPartition will see that message in the queue above, and checkpoint it
            // since the document was marked as corrupted.
            this.markAsCorrupt(error);
            this.q.resume();
        });
    }
    process(message) {
        if (this.closed) {
            return;
        }
        void this.q.push(message);
        this.updateActivityTime();
    }
    close(closeType) {
        if (this.closed) {
            return;
        }
        this.closed = true;
        // Stop any future processing
        this.q.kill();
        if (this.lambda) {
            this.lambda.close(closeType);
        }
        else {
            this.lambdaP.then((lambda) => {
                lambda.close(closeType);
            }, (error) => {
                // Lambda was never created - ignoring
            });
        }
    }
    isInactive(now = Date.now()) {
        return !this.context.hasPendingWork() && this.activityTimeoutTime && now > this.activityTimeoutTime;
    }
    /**
     * Marks this document partition as corrupt
     * Future messages will be checkpointed but no real processing will happen
     */
    markAsCorrupt(error, message) {
        var _a;
        this.corrupt = true;
        (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`Marking document as corrupted due to error: ${(0, util_1.inspect)(error)}`, {
            messageMetaData: {
                documentId: this.documentId,
                tenantId: this.tenantId,
            },
        });
        server_services_telemetry_1.Lumberjack.error(`Marking document as corrupted due to error`, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), error);
        this.context.error(error, { restart: false, tenantId: this.tenantId, documentId: this.documentId });
        if (message) {
            this.context.checkpoint(message);
        }
    }
    updateActivityTime() {
        this.activityTimeoutTime = Date.now() + this.activityTimeout;
    }
}
exports.DocumentPartition = DocumentPartition;
//# sourceMappingURL=documentPartition.js.map