"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Partition = void 0;
const events_1 = require("events");
const async_1 = require("async");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const checkpointManager_1 = require("./checkpointManager");
const context_1 = require("./context");
/**
 * Partition of a message stream. Manages routing messages to individual handlers. And then maintaining the
 * overall partition offset.
 */
class Partition extends events_1.EventEmitter {
    constructor(id, leaderEpoch, factory, consumer, logger) {
        super();
        this.id = id;
        this.logger = logger;
        this.closed = false;
        // Should we pass epoch with the context?
        const partitionConfig = { leaderEpoch };
        this.checkpointManager = new checkpointManager_1.CheckpointManager(id, consumer);
        this.context = new context_1.Context(this.checkpointManager, this.logger);
        this.context.on("error", (error, errorData) => {
            this.emit("error", error, errorData);
        });
        // Create the incoming message queue
        this.q = (0, async_1.queue)((message, callback) => {
            try {
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                const optionalPromise = this.lambda.handler(message);
                if (optionalPromise) {
                    optionalPromise
                        .then(callback)
                        .catch(callback);
                    return;
                }
                callback();
            }
            catch (error) {
                callback(error);
            }
        }, 1);
        this.q.pause();
        this.lambdaP = factory.create(partitionConfig, this.context);
        this.lambdaP.then((lambda) => {
            this.lambda = lambda;
            this.lambdaP = undefined;
            this.q.resume();
        }, (error) => {
            if (this.closed) {
                return;
            }
            const errorData = {
                restart: true,
            };
            this.emit("error", error, errorData);
            this.q.kill();
        });
        this.q.error((error) => {
            const errorData = {
                restart: true,
            };
            this.emit("error", error, errorData);
        });
    }
    process(rawMessage) {
        if (this.closed) {
            return;
        }
        void this.q.push(rawMessage);
    }
    close(closeType) {
        this.closed = true;
        // Stop any pending message processing
        this.q.kill();
        // Close checkpoint related classes
        this.checkpointManager.close();
        this.context.close();
        // Notify the lambda of the close
        if (this.lambda) {
            this.lambda.close(closeType);
            this.lambda = undefined;
        }
        else if (this.lambdaP) {
            // asynchronously close the lambda since it's not created yet
            this.lambdaP
                .then((lambda) => {
                lambda.close(closeType);
            }, (error) => {
                // Lambda never existed - no need to close
            })
                .finally(() => {
                this.lambda = undefined;
                this.lambdaP = undefined;
            });
        }
        this.removeAllListeners();
    }
    /**
     * Stops processing on the partition
     */
    async drain() {
        // Drain the queue of any pending operations
        const drainedP = new Promise((resolve, reject) => {
            var _a, _b;
            // If not entries in the queue we can exit immediatley
            if (this.q.length() === 0) {
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`No pending work for partition ${this.id}. Exiting early`);
                server_services_telemetry_1.Lumberjack.info(`No pending work for partition ${this.id}. Exiting early`);
                return resolve();
            }
            // Wait until the queue is drained
            (_b = this.logger) === null || _b === void 0 ? void 0 : _b.info(`Waiting for queue to drain for partition ${this.id}`);
            server_services_telemetry_1.Lumberjack.info(`Waiting for queue to drain for partition ${this.id}`);
            this.q.drain(() => {
                var _a;
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`Drained partition ${this.id}`);
                server_services_telemetry_1.Lumberjack.info(`Drained partition ${this.id}`);
                resolve();
            });
        });
        await drainedP;
        // Checkpoint at the latest offset
        await this.checkpointManager.flush();
    }
}
exports.Partition = Partition;
//# sourceMappingURL=partition.js.map