"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.PartitionManager = void 0;
const events_1 = require("events");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const partition_1 = require("./partition");
/**
 * The PartitionManager is responsible for maintaining a list of partitions for the given Kafka topic.
 * It will route incoming messages to the appropriate partition for the messages.
 */
class PartitionManager extends events_1.EventEmitter {
    constructor(factory, consumer, logger) {
        super();
        this.factory = factory;
        this.consumer = consumer;
        this.logger = logger;
        this.partitions = new Map();
        // Start rebalancing until we receive the first rebalanced message
        this.isRebalancing = true;
        this.stopped = false;
        // Place new Kafka messages into our processing queue
        this.consumer.on("data", (message) => {
            this.process(message);
        });
        this.consumer.on("rebalancing", (partitions) => {
            this.rebalancing(partitions);
        });
        this.consumer.on("rebalanced", (partitions) => {
            this.rebalanced(partitions);
        });
        this.consumer.on("error", (error, errorData) => {
            if (this.stopped) {
                return;
            }
            this.emit("error", error, errorData);
        });
    }
    async stop() {
        var _a;
        this.stopped = true;
        (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info("Stop requested");
        server_services_telemetry_1.Lumberjack.info("Stop requested");
        // Drain all pending messages from the partitions
        const partitionsStoppedP = [];
        for (const [, partition] of this.partitions) {
            const stopP = partition.drain();
            partitionsStoppedP.push(stopP);
        }
        await Promise.all(partitionsStoppedP);
        // Then stop them all
        for (const [, partition] of this.partitions) {
            partition.close(server_services_core_1.LambdaCloseType.Stop);
        }
        this.partitions.clear();
        this.removeAllListeners();
    }
    process(message) {
        var _a;
        if (this.stopped) {
            return;
        }
        if (this.isRebalancing) {
            (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`Ignoring ${message.topic}:${message.partition}@${message.offset} due to pending rebalance`);
            server_services_telemetry_1.Lumberjack.info(`Ignoring ${message.topic}:${message.partition}@${message.offset} due to pending rebalance`);
            return;
        }
        const partition = this.partitions.get(message.partition);
        if (!partition) {
            this.emit("error", `Received message for untracked partition ${message.topic}:${message.partition}@${message.offset}`);
            return;
        }
        partition.process(message);
    }
    /**
     * Called when rebalancing starts
     * Note: The consumer may decide to only emit "rebalanced" if it wants to skip closing existing partitions
     * @param partitions - Assigned partitions before the rebalance
     */
    rebalancing(partitions) {
        var _a, _b;
        (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`Rebalancing partitions: ${JSON.stringify(partitions)}`);
        server_services_telemetry_1.Lumberjack.info(`Rebalancing partitions: ${JSON.stringify(partitions)}`);
        this.isRebalancing = true;
        for (const [id, partition] of this.partitions) {
            (_b = this.logger) === null || _b === void 0 ? void 0 : _b.info(`Closing partition ${id} due to rebalancing`);
            server_services_telemetry_1.Lumberjack.info(`Closing partition ${id} due to rebalancing`);
            partition.close(server_services_core_1.LambdaCloseType.Rebalance);
        }
        this.partitions.clear();
    }
    /**
     * Called when rebalanced occurs
     * @param partitions - Assigned partitions after the rebalance.
     * May contain partitions that have been previously assigned to this consumer
     */
    rebalanced(partitions) {
        var _a, _b;
        if (this.stopped) {
            return;
        }
        this.isRebalancing = false;
        const partitionsMap = new Map(partitions.map((partition) => [partition.partition, partition]));
        // close and remove existing partitions that are no longer assigned
        const existingPartitions = Array.from(this.partitions);
        for (const [id, partition] of existingPartitions) {
            if (!partitionsMap.has(id)) {
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`Closing partition ${id} due to rebalancing`);
                server_services_telemetry_1.Lumberjack.info(`Closing partition ${id} due to rebalancing`);
                partition.close(server_services_core_1.LambdaCloseType.Rebalance);
                this.partitions.delete(id);
            }
        }
        // create new partitions
        for (const partition of partitions) {
            if (this.partitions.has(partition.partition)) {
                // this partition already exists
                // it must have existed before the rebalance
                continue;
            }
            // eslint-disable-next-line max-len
            (_b = this.logger) === null || _b === void 0 ? void 0 : _b.info(`Creating ${partition.topic}: Partition ${partition.partition}, Epoch ${partition.leaderEpoch}, Offset ${partition.offset} due to rebalance`);
            // eslint-disable-next-line max-len
            server_services_telemetry_1.Lumberjack.info(`Creating ${partition.topic}: Partition ${partition.partition}, Epoch ${partition.leaderEpoch}, Offset ${partition.offset} due to rebalance`);
            const newPartition = new partition_1.Partition(partition.partition, partition.leaderEpoch, this.factory, this.consumer, this.logger);
            // Listen for error events to know when the partition has stopped processing due to an error
            newPartition.on("error", (error, errorData) => {
                if (this.stopped) {
                    return;
                }
                this.emit("error", error, errorData);
            });
            this.partitions.set(partition.partition, newPartition);
        }
    }
}
exports.PartitionManager = PartitionManager;
//# sourceMappingURL=partitionManager.js.map