"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.KafkaRunner = void 0;
const util_1 = require("util");
const common_utils_1 = require("@fluidframework/common-utils");
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const partitionManager_1 = require("./partitionManager");
class KafkaRunner {
    constructor(factory, consumer) {
        this.factory = factory;
        this.consumer = consumer;
        this.stopped = false;
        this.runnerMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.KafkaRunner);
    }
    // eslint-disable-next-line @typescript-eslint/promise-function-async
    start(logger) {
        if (this.deferred) {
            this.runnerMetric.error("Runner already started");
            throw new Error("Already started");
        }
        const deferred = new common_utils_1.Deferred();
        this.deferred = deferred;
        process.on("warning", (msg) => {
            console.trace("Warning", msg);
        });
        this.factory.on("error", (error) => {
            this.runnerMetric.error("Kafka factory encountered an error", error);
            deferred.reject(error);
        });
        this.partitionManager = new partitionManager_1.PartitionManager(this.factory, this.consumer, logger);
        this.partitionManager.on("error", (error, errorData) => {
            var _a, _b;
            const documentId = (_a = errorData === null || errorData === void 0 ? void 0 : errorData.documentId) !== null && _a !== void 0 ? _a : "";
            const tenantId = (_b = errorData === null || errorData === void 0 ? void 0 : errorData.tenantId) !== null && _b !== void 0 ? _b : "";
            const lumberProperties = (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId);
            const metadata = {
                messageMetaData: {
                    documentId,
                    tenantId,
                },
            };
            this.runnerMetric.setProperties(lumberProperties);
            if (errorData && !errorData.restart) {
                const errorMsg = "KakfaRunner encountered an error that is not configured to trigger restart";
                logger === null || logger === void 0 ? void 0 : logger.error(errorMsg, metadata);
                logger === null || logger === void 0 ? void 0 : logger.error((0, util_1.inspect)(error), metadata);
                if (!this.runnerMetric.isCompleted()) {
                    this.runnerMetric.error(errorMsg, error);
                }
                else {
                    server_services_telemetry_1.Lumberjack.error(errorMsg, lumberProperties, error);
                }
            }
            else {
                const errorMsg = "KakfaRunner encountered an error that will trigger a restart";
                logger === null || logger === void 0 ? void 0 : logger.error(errorMsg, metadata);
                logger === null || logger === void 0 ? void 0 : logger.error((0, util_1.inspect)(error), metadata);
                if (!this.runnerMetric.isCompleted()) {
                    this.runnerMetric.error(errorMsg, error);
                }
                else {
                    server_services_telemetry_1.Lumberjack.error(errorMsg, lumberProperties, error);
                }
                deferred.reject(error);
            }
        });
        this.stopped = false;
        return deferred.promise;
    }
    /**
     * Signals to stop the service
     */
    async stop() {
        var _a;
        if (!this.deferred || this.stopped) {
            return;
        }
        this.stopped = true;
        // Stop listening for new updates
        await this.consumer.pause();
        // Stop the partition manager
        await ((_a = this.partitionManager) === null || _a === void 0 ? void 0 : _a.stop());
        // Dispose the factory
        await this.factory.dispose();
        // Close the underlying consumer, but setting a timeout for safety
        await (0, server_services_client_1.promiseTimeout)(30000, this.consumer.close());
        // Mark ourselves done once the partition manager has stopped
        this.deferred.resolve();
        this.deferred = undefined;
        this.runnerMetric.success("Kafka runner stopped");
    }
}
exports.KafkaRunner = KafkaRunner;
//# sourceMappingURL=runner.js.map